# This is the initialization module which contains the primitive stuff which 
# are needed by the db_models package to work. It contains the following:
# 1. Base : class
#       It is the base class for all the db model that will be going to
#       present in the db_models. This class will serve the purpose
#       of providing the common functionalities.
# 2. session : sqlalchemy Object
#       This reference variable will contain the Session() object after initialization.
#       The session is used to interact with the database. Seperating out the session
#       ensures that multiple session does not form during the execution.


from builtins import object
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from db_paribhasha import db_constants
from db_paribhasha import db_creds
from db_paribhasha.base_models.get_secret import get_dbsecret

class DatabaseSession(object):
    def database_url(self):
        '''
        This function prepares and returns the database url for the
        connection preparation with the postgres database using sqlalchemy.
    
        '''
        database_user = db_creds.DATABASE_USER
        database_host = db_creds.DATABASE_HOST
        database_name = db_creds.DATABASE_NAME
        db_password = eval(get_dbsecret()).get("password")

        dsn_tempate = 'postgresql+psycopg2://{usr}:{pass}@{host}/{db_name}'
        format_args = {'usr':database_user,
                       'pass':db_password,
                       'host':database_host,
                       'db_name':database_name
                      }
        db_url = dsn_tempate.format(**format_args)

        return db_url

    def initiate_session(self):
        '''
        This function initializes the database session. This session is the
        object of Session() which acts a session factory to get the session(connection)
        with the database. 
        '''
        db_url =  self.database_url()
        engine = create_engine(db_url)
        Session = sessionmaker(bind=engine)
        session = Session()
        return engine,session
    
    __db_session = None
    __db_engine = None

    def __init__(self):
        if DatabaseSession.__db_session is None:
            engine, session = self.initiate_session()
            DatabaseSession.__db_session = session
            DatabaseSession.__db_engine = engine

    @staticmethod
    def dispose_session():
        DatabaseSession.__db_session.close()
        DatabaseSession.__db_engine.dispose()

    @property
    def db_session(self):
        return DatabaseSession.__db_session


session = DatabaseSession().db_session

BaseModel = declarative_base()

class Base(BaseModel):
    '''  
      It is an abstract class.It is the base class for all the 
      db model that will be going to present in the db_models. 
      This class will serve the purpose of providing the common 
      functionalities.
    '''
    __abstract__ = True

    @classmethod
    def get_object(cls,id):
        '''
        This is the factory method which return the object of the class
        whose row is present in the database table with the given id

        PARAMETERS
        ----------
        cls : type
            It is object of current class definition with whose reference
            the method is called.
        id : int|str
            primary key id of the database model whose corresponding
            ORM object is to be returned
        
        RETURNS
        --------
        object:
            Object of Base and/or its subclasses

        '''
        model_obj =  session.query(cls).get(id)
        if model_obj is None:
            raise LookupError(db_constants.INVALID_ID.format(cls.__table__))

        return model_obj

    @classmethod
    def get_object_with_ids(cls,ids):
        '''
        This functions returns the list of objects of the db model with the ids
        in the list given.

        PARAMETERS
        ----------
        cls : type
            It is object of current class definition with whose reference
            the method is called.
        ids : list
            list of primary key ids of the database model whose corresponding
            ORM object is to be returned
        
        RETURNS
        --------
        object:
            list of Object of Base and/or its subclasses

        '''
        objects = session.query(cls).\
        filter(cls.id.in_(ids)).all()

        object_dict = { obj.id:obj for obj in objects}
        ordered_list =list()
        for id in ids:
            ordered_list.append(object_dict[id])
        return ordered_list

    
    def save_to_db(self):
        """
        Saves data in the database
        """
        session.add(self)
        session.commit()
        
    
    def __str__(self):
        return self.__dict__
